==> GotoBLAS-1.12.tar.gz: GotoBLAS/lapack/getrf/getrf_parallel.c <==
/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dm1 = -1.;

#if defined(CORE_PPC970) || defined(CORE_POWER4)
#define NB_OFFSET 64
#endif

#ifdef PENTIUM4
#define NB_OFFSET 80
#endif

#ifdef ITANIUM2
#define NB_OFFSET 32
#endif

#ifndef NB_OFFSET
#define NB_OFFSET (jb / 2)
#endif

static void inner_thread(long m, long n, long jb, FLOAT dummy1,
#ifdef COMPLEX
			 FLOAT dummy2,
#endif
			 FLOAT *offsetA, long lda, FLOAT *offsetB, long offset,
			 blasint *ipiv, long dummy3, void *buffer){

  long is, imin, js, jmin, jc, jcmin;
  FLOAT *offsetC;

  FLOAT *sa = (FLOAT *)((void *)buffer + GEMM_BUFFER_A_OFFSET);
  FLOAT *sb = (FLOAT *)((void *)buffer + GEMM_BUFFER_B_OFFSET
			+ ((GEMM_P * GEMM_Q * COMPSIZE * SIZE +
			    GEMM_BUFFER_A_OFFSET + GEMM_ALIGN) & ~GEMM_ALIGN));
  
  for (js = 0; js < n; js += GEMM_R){
    jmin = n - js;
    if (jmin > GEMM_R) jmin = GEMM_R;
    
    offsetC = offsetB + js * lda * COMPSIZE;
    
    LASWP_PLUS(jmin, offset + 1, offset + jb, ZERO, 
#ifdef COMPLEX
	       ZERO,
#endif
	       offsetC - offset * COMPSIZE, lda, NULL, 0 , ipiv, 1);
    
    GEMM_ONCOPY (jb, jmin, offsetC, lda, sb);

    for (jc = 0; jc < jb; jc += GEMM_P) {
      jcmin = jb - jc;
      if (jcmin > GEMM_P) jcmin = GEMM_P;
      
      TRSM_ILTCOPY(jb, jcmin, offsetA + jc * COMPSIZE, lda, jc, sa);
      
      TRSM_KERNEL_LT(jcmin, jmin, jb, dm1,
#ifdef COMPLEX
		     ZERO,
#endif
		     sa, sb, offsetC + jc * COMPSIZE, lda, jc);
    }
    
    for (is = 0; is < m; is += GEMM_P){
      imin = m - is;
      if (imin > GEMM_P) imin = GEMM_P;
      
      GEMM_ITCOPY (jb, imin, offsetA + (is + jb) * COMPSIZE, lda, sa);
      
      GEMM_KERNEL_N(imin, jmin, jb, dm1,
#ifdef COMPLEX
		    ZERO,
#endif
		    sa, sb, offsetC + (is + jb) * COMPSIZE, lda);
    }
  }
}

static blasint getrf_parallel(long m, long n, FLOAT *a, long lda, 
			  blasint *ipiv, blas_queue_t *queue, long offset, FLOAT *buffer){

  long i, j, blocking;
  blasint iinfo, info;
  long jb, min_mn;
  FLOAT *offsetA, *offsetB;
  long execflag;
  long width, oldjb;
  long num_cpu;

  if (m == 0 || n == 0) return 0;

  if (n <= DTB_ENTRIES) {
    info = GETRF_SINGLE(m, n, a, lda, ipiv, offset, buffer);
    return info;
  }

  min_mn = MIN(m, n);

  if (n > GEMM_Q) {
    blocking = GEMM_Q;
  } else {
    if (n & 1) {
      blocking = (n / 2) + 1;
    } else {
      blocking = (n / 2);
    }
    if (blocking <= GEMM_UNROLL_N) blocking = GEMM_UNROLL_N; 
  }

  info = 0;
  execflag = 0;

  jb = MIN(min_mn, blocking);

  for (j = 0; j < min_mn; j += oldjb) {
    
    offsetA = a +  (j + (j     ) * lda) * COMPSIZE;
    offsetB = a +  (j + (j + jb) * lda) * COMPSIZE;

    if ((j == 0) && (jb > GEMM_UNROLL_N)) {
      iinfo = getrf_parallel(m - j, jb, offsetA, lda, ipiv, queue, j + offset, buffer);
    } else {
      iinfo = GETRF_SINGLE  (m - j, jb, offsetA, lda,ipiv, j + offset, buffer);
    }

    if (iinfo && !info) info = iinfo + j;

    oldjb = jb;

    i = n - j - jb;

    if (i > 0) {

      /* Adjusting MAIN CPU task */
      if (i >= blocking * ( 2 * blas_cpu_number - 1)) {
	/* we can use fixed blocking size */

	width  = ((blas_quickdivide(i + jb + NB_OFFSET, blas_cpu_number)
		   + GEMM_UNROLL_N - 1) & ~(GEMM_UNROLL_N - 1)) - jb;
	jb = blocking;
      } else {
	/* rest is too small */
	width  = blas_quickdivide(i, 2 * blas_cpu_number - 1);
	if (width <= 0) width = 1;
	jb = width;
      }
      
      queue[0].m   = m - j - oldjb;
      queue[0].n   = width;
      queue[0].k   = oldjb;
      queue[0].a   = (void *)offsetA;
      queue[0].lda = lda;
      queue[0].b   = (void *)offsetB;
      queue[0].ldb = j + offset;
      queue[0].c   = (void *)ipiv;
      queue[0].ldc = 0;
      
      offsetB   += lda * width * COMPSIZE;
      i         -= width;
      
      /* Determing other CPU task */
      num_cpu = 1;
      
      if (execflag){
	exec_blas_async_wait(&queue[1]);
	execflag = 0;
      }
      
      while (i > 0){
	
	width  = (blas_quickdivide(i, blas_cpu_number - num_cpu));
	
	queue[num_cpu].m      = m - j - oldjb;
	queue[num_cpu].n      = width;
	queue[num_cpu].k      = oldjb;
	queue[num_cpu].a      = (void *)offsetA;
	queue[num_cpu].lda    = lda;
	queue[num_cpu].b      = (void *)offsetB;
	queue[num_cpu].ldb    = j + offset;
	queue[num_cpu].c      = (void *)ipiv;

	queue[num_cpu].status = BLAS_STATUS_NOTYET;
	queue[num_cpu].next   = &queue[num_cpu + 1];
	queue[num_cpu].last   = 0;
	
	offsetB += lda * width * COMPSIZE;
	i       -= width;
	
	num_cpu ++;
      }
      
      queue[num_cpu - 1].next  = NULL;
      queue[num_cpu - 1].last = 1;
      
      if (num_cpu > 1) {
	exec_blas_async(&queue[1]);
	execflag = 1;
      }
      
      inner_thread(queue[0].m, queue[0].n, queue[0].k, ZERO,
#ifdef COMPLEX
		   ZERO,
#endif
		   queue[0].a, queue[0].lda, 
		   queue[0].b, queue[0].ldb,
		   queue[0].c, queue[0].ldc, buffer);
      
    }
  }

  if (execflag){
    exec_blas_async_wait(&queue[1]);
    execflag = 0;
  }
      
#ifndef COMPLEX
#ifdef XDOUBLE
  int mode  =  BLAS_XDOUBLE | BLAS_REAL;
#elif defined(DOUBLE)
  int mode  =  BLAS_DOUBLE  | BLAS_REAL;
#else
  int mode  =  BLAS_SINGLE  | BLAS_REAL;
#endif  
#else
#ifdef XDOUBLE
  int mode  =  BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
  int mode  =  BLAS_DOUBLE  | BLAS_COMPLEX;
#else
  int mode  =  BLAS_SINGLE  | BLAS_COMPLEX;
#endif  
#endif

  jb = MIN(min_mn, blocking);
    
  for (j = 0; j < min_mn; j += oldjb) {

    blas_level1_thread(mode, jb, j + jb + 1 + offset, min_mn + offset, ZERO, ZERO, 
		       a + (- offset + j * lda) * COMPSIZE, lda, 
		       NULL, 0, ipiv, 1, (void *)LASWP_PLUS);

    oldjb = jb;

    i = n - j - jb;

    if (i >= blocking * ( 2 * blas_cpu_number - 1)) {
      jb = blocking;
    } else {
      jb = blas_quickdivide(i, 2 * blas_cpu_number - 1);
      if (jb <= 0) jb = 1;
    }
  }

  return info;
}

blasint CNAME(long m, long n, FLOAT *a, long lda, blasint *ipiv, long dummy, FLOAT *buffer){

  long i;
  blasint info;
  blas_queue_t queue[MAX_CPU_NUMBER];

  if (m == 0 || n == 0) return 0;

  info = 0;

  if(blas_cpu_number == 1) {

    info = GETRF_SINGLE(m, n, a, lda, ipiv, 0, buffer);

  } else {
    for (i = 0; i < blas_cpu_number; i++){

      blas_queue_init(&queue[i]);
      
#ifndef COMPLEX
#ifdef XDOUBLE
      queue[i].mode = BLAS_XDOUBLE | BLAS_REAL;
#elif defined(DOUBLE)
      queue[i].mode = BLAS_DOUBLE  | BLAS_REAL;
#else
      queue[i].mode = BLAS_SINGLE  | BLAS_REAL;
#endif  
#else
#ifdef XDOUBLE
      queue[i].mode = BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
      queue[i].mode = BLAS_DOUBLE  | BLAS_COMPLEX;
#else
      queue[i].mode = BLAS_SINGLE  | BLAS_COMPLEX;
#endif  
#endif
      
      queue[i].routine = (void *)inner_thread;
      queue[i].status = BLAS_STATUS_FINISHED;
      queue[i].last = 1;
    }
    
    info = getrf_parallel(m, n, a, lda, ipiv, queue, 0, buffer);
    
    for (i = 0; i < blas_cpu_number; i++) blas_queue_destroy(&queue[i]);
  }

  return info;
}

==> GotoBLAS-1.15.tar.gz: GotoBLAS/lapack/getrf/getrf_parallel.c <==
/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include "common.h"

static FLOAT dm1 = -1.;

#undef TIMING

#ifdef TIMING
extern long exit_time[];
#endif

#ifndef FACTOR
#define FACTOR 0.66666667
#endif

static long calc_width(long x, long y, long a, long n){
  return ((x + a) * y + FACTOR * x * a * (1 - blas_cpu_number)) / ( blas_cpu_number * (x + a));
}

static void inner_thread(long m, long n, long jb, FLOAT dummy1,
#ifdef COMPLEX
			 FLOAT dummy2,
#endif
			 FLOAT *offsetA, long lda, FLOAT *offsetB, long offset,
			 blasint *ipiv, long dummy3, FLOAT *sa, FLOAT *sb){

  long is, imin, js, jmin, jc, jcmin;
  FLOAT *offsetC;

  for (js = 0; js < n; js += GEMM_R){
    jmin = n - js;
    if (jmin > GEMM_R) jmin = GEMM_R;
    
    offsetC = offsetB + js * lda * COMPSIZE;
    
#if GEMM_UNROLL_N > 8
    LASWP_PLUS(jmin, offset + 1, offset + jb, ZERO, 
#ifdef COMPLEX
	       ZERO,
#endif
	       offsetC - offset * COMPSIZE, lda, NULL, 0 , ipiv, 1);
    
    GEMM_ONCOPY (jb, jmin, offsetC, lda, sb);
#else
    LASWP_NCOPY(jmin, offset + 1, offset + jb,
		offsetC - offset * COMPSIZE, lda, ipiv, sb);
#endif

    for (jc = 0; jc < jb; jc += GEMM_P) {
      jcmin = jb - jc;
      if (jcmin > GEMM_P) jcmin = GEMM_P;
      
      TRSM_ILTCOPY(jb, jcmin, offsetA + jc * COMPSIZE, lda, jc, sa);
      
      TRSM_KERNEL_LT(jcmin, jmin, jb, dm1,
#ifdef COMPLEX
		     ZERO,
#endif
		     sa, sb, offsetC + jc * COMPSIZE, lda, jc);
    }
    
    for (is = 0; is < m; is += GEMM_P){
      imin = m - is;
      if (imin > GEMM_P) imin = GEMM_P;
      
      GEMM_ITCOPY (jb, imin, offsetA + (is + jb) * COMPSIZE, lda, sa);
      
      GEMM_KERNEL_N(imin, jmin, jb, dm1,
#ifdef COMPLEX
		    ZERO,
#endif
		    sa, sb, offsetC + (is + jb) * COMPSIZE, lda);
    }
  }
}

static blasint getrf_parallel(long m, long n, FLOAT *a, long lda, 
			  blasint *ipiv, blas_queue_t *queue, long offset, FLOAT *buffer){

  long i, j, blocking;
  blasint iinfo, info;
  long jb, min_mn;
  FLOAT *offsetA, *offsetB;
  long execflag;
  long width, oldjb;
  long num_cpu;
#ifdef TIMING
  long current_time;
#endif

  FLOAT *sa = (FLOAT *)((void *)buffer + GEMM_BUFFER_A_OFFSET);
  FLOAT *sb = (FLOAT *)((void *)buffer + GEMM_BUFFER_B_OFFSET
			+ ((GEMM_P * GEMM_Q * COMPSIZE * SIZE +
			    GEMM_BUFFER_A_OFFSET + GEMM_ALIGN) & ~GEMM_ALIGN));
  

  if (m == 0 || n == 0) return 0;

  if (n <= DTB_ENTRIES) {
    info = GETRF_SINGLE(m, n, a, lda, ipiv, offset, sa, sb);
    return info;
  }

  min_mn = MIN(m, n);

  blocking = n / 4;
  if (blocking > GEMM_Q) blocking = GEMM_Q;

  info = 0;
  execflag = 0;

  jb = MIN(min_mn, blocking);

  for (j = 0; j < min_mn; j += oldjb) {
    
    offsetA = a +  (j + (j     ) * lda) * COMPSIZE;
    offsetB = a +  (j + (j + jb) * lda) * COMPSIZE;

#if 1
    if ((j == 0) && (jb > GEMM_UNROLL_N * 2) && (offset == 0)) {
      iinfo = getrf_parallel(m - j, jb, offsetA, lda, ipiv, queue, j + offset, buffer);
    } else {
      iinfo = GETRF_SINGLE  (m - j, jb, offsetA, lda,ipiv, j + offset, sa, sb);
    }
#else
      iinfo = GETRF_SINGLE  (m - j, jb, offsetA, lda,ipiv, j + offset, sa, sb);
#endif

    if (iinfo && !info) info = iinfo + j;

    oldjb = jb;

    i = n - j - jb;

    if (i > 0) {
      width = calc_width(m - j - jb, n - j - jb, jb, blas_cpu_number);

      if (width > i) width = i;
      
      if (width < jb) {
	width  = blas_quickdivide(i, 2 * blas_cpu_number - 1);
	if (width <= 0) width = 1;
	jb = width;
      }

      queue[0].m   = m - j - oldjb;
      queue[0].n   = width;
      queue[0].k   = oldjb;
      queue[0].a   = (void *)offsetA;
      queue[0].lda = lda;
      queue[0].b   = (void *)offsetB;
      queue[0].ldb = j + offset;
      queue[0].c   = (void *)ipiv;
      queue[0].ldc = 0;
      
      offsetB   += lda * width * COMPSIZE;
      i         -= width;
      
      /* Determing other CPU task */
      num_cpu = 1;
      
#ifdef TIMING
      current_time = rpcc();
#endif
      if (execflag){
	exec_blas_async_wait(&queue[1]);
	execflag = 0;
      }

#ifdef TIMING
       {
	 int xxx;
	 if (j > 0) {
	   for (xxx = 0; xxx < blas_cpu_number - 1; xxx++){
	     fprintf(stderr, "%12ld ", current_time - exit_time[xxx]);
	   }
	   fprintf(stderr, "\n");
	 }
       }
#endif

      while (i > 0){
	
	width  = (blas_quickdivide(i, blas_cpu_number - num_cpu));
	
	queue[num_cpu].m      = m - j - oldjb;
	queue[num_cpu].n      = width;
	queue[num_cpu].k      = oldjb;
	queue[num_cpu].a      = (void *)offsetA;
	queue[num_cpu].lda    = lda;
	queue[num_cpu].b      = (void *)offsetB;
	queue[num_cpu].ldb    = j + offset;
	queue[num_cpu].c      = (void *)ipiv;

	queue[num_cpu].status = BLAS_STATUS_NOTYET;
	queue[num_cpu].next   = &queue[num_cpu + 1];
	queue[num_cpu].last   = 0;
	
	offsetB += lda * width * COMPSIZE;
	i       -= width;
	
	num_cpu ++;
      }
      
      queue[num_cpu - 1].next  = NULL;
      queue[num_cpu - 1].last = 1;
      
      if (num_cpu > 1) {
	exec_blas_async(&queue[1]);
	execflag = 1;
      }
      
      inner_thread(queue[0].m, queue[0].n, queue[0].k, ZERO,
#ifdef COMPLEX
		   ZERO,
#endif
		   queue[0].a, queue[0].lda, 
		   queue[0].b, queue[0].ldb,
		   queue[0].c, queue[0].ldc, sa, sb);
      
    }
  }

  if (execflag){
    exec_blas_async_wait(&queue[1]);
    execflag = 0;
  }
      
#ifndef COMPLEX
#ifdef XDOUBLE
  int mode  =  BLAS_XDOUBLE | BLAS_REAL;
#elif defined(DOUBLE)
  int mode  =  BLAS_DOUBLE  | BLAS_REAL;
#else
  int mode  =  BLAS_SINGLE  | BLAS_REAL;
#endif  
#else
#ifdef XDOUBLE
  int mode  =  BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
  int mode  =  BLAS_DOUBLE  | BLAS_COMPLEX;
#else
  int mode  =  BLAS_SINGLE  | BLAS_COMPLEX;
#endif  
#endif

  jb = MIN(min_mn, blocking);
    
  for (j = 0; j < min_mn; j += oldjb) {
    blas_level1_thread(mode, jb, j + jb + 1 + offset, min_mn + offset, ZERO, ZERO, 
		       a + (- offset + j * lda) * COMPSIZE, lda, 
		       NULL, 0, ipiv, 1, (void *)LASWP_PLUS);

    oldjb = jb;

    i = n - j - jb;

    width = calc_width(m - j - jb, n - j - jb, jb, blas_cpu_number);

    if (width > i) width = i;
    
    if (width < jb) {
      width  = blas_quickdivide(i, 2 * blas_cpu_number - 1);
      if (width <= 0) width = 1;
      jb = width;
    }
  }

  return info;
}

blasint CNAME(long m, long n, FLOAT *a, long lda, blasint *ipiv, long dummy, FLOAT *buffer){

  long i;
  blasint info;
  blas_queue_t queue[MAX_CPU_NUMBER];

  if (m == 0 || n == 0) return 0;

  info = 0;

  for (i = 0; i < blas_cpu_number; i++){
    
    blas_queue_init(&queue[i]);
    
#ifndef COMPLEX
#ifdef XDOUBLE
    queue[i].mode = BLAS_XDOUBLE | BLAS_REAL;
#elif defined(DOUBLE)
    queue[i].mode = BLAS_DOUBLE  | BLAS_REAL;
#else
    queue[i].mode = BLAS_SINGLE  | BLAS_REAL;
#endif  
#else
#ifdef XDOUBLE
    queue[i].mode = BLAS_XDOUBLE | BLAS_COMPLEX;
#elif defined(DOUBLE)
    queue[i].mode = BLAS_DOUBLE  | BLAS_COMPLEX;
#else
    queue[i].mode = BLAS_SINGLE  | BLAS_COMPLEX;
#endif  
#endif
    
    queue[i].routine = (void *)inner_thread;
    queue[i].status = BLAS_STATUS_FINISHED;
    queue[i].last = 1;
  }
  
  info = getrf_parallel(m, n, a, lda, ipiv, queue, 0, buffer);
  
  for (i = 0; i < blas_cpu_number; i++) blas_queue_destroy(&queue[i]);

  return info;
}
